/*+------------------------------------------------------------------+
//| 1:39 PM 6/7/2020                         indicator_utilities.mqh |
//|                             Copyright  2020, WHRoeder@yahoo.com |
//|                                        mailto:WHRoeder@yahoo.com |
//+------------------------------------------------------------------+
//| Indicator support.                                               |
//| A) Add inputs and buffer definitions (as outlined in             |
//|    indicator_buffers.mqh and include THAT FILE AFTER all buffer  |
//|    #define's (or at end of the main file) to define              |
//|    OnInitBuffers().                                              |
//| B) To use AlignableArrow,                                        |
//|    1 Choose which AlignableArrow (AA) you want, either a         |
//|      constant or a input variable.                               |
//|    2 Set buffer's arrow code using arr(AA)                       |
//|    3 From OnTimer() call OnTimerAlignable()                      |
//|    4 Adjust PRICEs using gsAAA.align_top(AA, p)                  |
//+------------------------------------------------------------------+*/
#include    <common_utilities.mqh>  // INDEX, COUNT, PRICE, gsTicker

//{ Special Arrow codes for OBJECT_ARROW that exactly points to price and time.
// These cannot be used in custom indicator's DRAW_ARROW buffers.
enum SpecialArrowCode{
   eSAC_ARROW_UR=1,     // Upwards arrow with tip rightwards
   eSAC_ARROW_DR,       // Downwards arrow with tip rightwards
   eSAC_TRIANGLE_LEFT,  // Left pointing triangle
   eSAC_DASH,           // Dash symbol
   eSAC_LEFT_PRICE,     // Left sided price label  // SYMBOL_LEFTPRICE
   eSAC_RIGHT_PRICE};   // Right sided price label // SYMBOL_RIGHTPRICE
//} Special Arrow codes

/* Wingding codes for OBJECT_ARROW and custom indicator's DRAW_ARROW buffers.
 * They do not point exactly at price, see below.
 * See: "Wingdings - Objects Constants - Constants, Enumerations and Structures
 * - MQL4 Reference"  //docs.mql4.com/constants/objectconstants/wingdings
 * and: "Names from Wingdings character set and equivalent Unicode characters" 
 * //www.alanwood.net/demos/wingdings.html */
enum  Wingding {              eWD_THUMBS_UP=67,          eWD_THUMBS_DOWN,
eWD_CIRCLE_SOLID=108,         eWD_SQUARE_SOLID=110,      eWD_BOX,
eWD_DIAMOND_SOLID_SMALL=115,  eWD_DIAMOND_SOLID,         eWD_RHOMBUS_SOLID,
eWD_RHOMBUS_EMPTY_X,          eWD_RHOMBUS_SOLID_SMALL,   eWD_INTEREST=122,
eWD_CIRCLED_0=128,            eWD_CIRCLED_1,             eWD_CIRCLED_2,
eWD_CIRCLED_3,                eWD_CIRCLED_4,             eWD_CIRCLED_5,
eWD_CIRCLED_6,                eWD_CIRCLED_7,             eWD_CIRCLED_8,
eWD_CIRCLED_9,                eWD_CIRCLED_10,            eWD_CIRCLED_EMPTY_0,
eWD_CIRCLED_EMPTY_1,          eWD_CIRCLED_EMPTY_2,       eWD_CIRCLED_EMPTY_3,
eWD_CIRCLED_EMPTY_4,          eWD_CIRCLED_EMPTY_5,       eWD_CIRCLED_EMPTY_6,
eWD_CIRCLED_EMPTY_7,          eWD_CIRCLED_EMPTY_8,       eWD_CIRCLED_EMPTY_9,
eWD_CIRCLED_EMPTY_10,        eWD_CIRCLE_SOLID_SMALL=158,eWD_CIRCLE_SOLID_MEDIUM,
eWD_CIRCLE=161,               eWD_CIRCLE_BOLD,           eWD_CIRCLE_HEAVY,
eWD_FISHEYE,                  eWD_BULLSEYE,          eWD_SQUARE_SOLID_SMALL=167,
eWD_STAR_3=169,               eWD_STAR_4,                eWD_STAR_5,
eWD_STAR_6,                   eWD_STAR_8,                eWD_STAR_12,
eWD_POSITION_SQUARE=176,      eWD_POSITION_CIRCLE,       eWD_DIAMOND_CONCAVE,
eWD_CIRCLED_PENTASTAR=181,    eWD_OCLOCK_1=183,          eWD_OCLOCK_2,
eWD_OCLOCK_3,                 eWD_OCLOCK_4,              eWD_OCLOCK_5,
eWD_OCLOCK_6,                 eWD_OCLOCK_7,              eWD_OCLOCK_8,
eWD_OCLOCK_9,                 eWD_OCLOCK_10,             eWD_OCLOCK_11,
eWD_OCLOCK_12,                eWD_CHEVRON_RIGHT=216,     eWD_CHEVRON_UP,
eWD_CHEVRON_DOWN,             eWD_ARROW_EMPTY_CIRCLED_RIGHT=220,
eWD_ARROW_EMPTY_CIRCLED_UP, eWD_ARROW_EMPTY_CIRCLED_DOWN,eWD_ARROW_RIGHT=224,
eWD_ARROW_UP,                 eWD_ARROW_DOWN,            eWD_ARROW_UR=228,
eWD_ARROW_DR=230,             eWD_ARROW_BOLD_RIGHT=232,  eWD_ARROW_BOLD_UP,
eWD_ARROW_BOLD_DOWN,          eWD_ARROW_BOLD_UR=236,     eWD_ARROW_BOLD_DR=238,
eWD_ARROW_EMPTY_RIGHT=240,    eWD_ARROW_EMPTY_UP,        eWD_ARROW_EMPTY_DOWN,
eWD_ARROW_EMPTY_UR=246,       eWD_ARROW_EMPTY_DR=248,    eWD_X_MARK=251,
eWD_CHECK_MARK,               eWD_X_MARK_BOX,            eWD_CHECK_MARK_BOX   };

//{ Exactly alignable arrows.
enum AlignableArrow{					// Exactly Alignable arrows.
   ePA_CIRCLE_SOLID,                      // Circle Solid Large
   ePA_CIRCLE_SOLID_MEDIUM,               // Circle Solid Medium
   ePA_SQUARE_SOLID,                      // Square Large
   ePA_SQUARE_SOLID_SMALL,                // Square Small
   ePA_BOX,                               // Box
   ePA_DIAMOND_SOLID,                     // Diamond Solid
   ePA_RHOMBUS_SOLID,                     // Rhombus Large
   ePA_RHOMBUS_EMPTY_X,                   // Rhombus Large w/ white X
   ePA_INTEREST,                          // Interest Point
   ePA_CIRCLE,                            // Circle Large
   ePA_CIRCLE_BOLD,                       // Circle Large Bold
   ePA_CIRCLE_HEAVY,                      // Circle Large Heavy
   ePA_FISHEYE,                           // Fish Eye
   ePA_BULLSEYE,                          // Bulls Eye
   ePA_STAR_3,                            // 3 point Star
   ePA_STAR_4,                            // 4 point Star
   ePA_STAR_5,                            // 5 point Star
   ePA_STAR_6,                            // 6 point Star
   ePA_STAR_8,                            // 8 point Star
   ePA_STAR_12,                           // 12 point Star
   ePA_POSITION_SQUARE,                   // Position indicator Square
   ePA_POSITION_CIRCLE,                   // Position indicator Circle
   ePA_DIAMOND_CONCAVE,                   // Diamond Empty Concave
   ePA_CIRCLED_PENTASTAR,                 // 5 point star Circled
   ePA_CHEVRON_RIGHT,                     // Chevron Right
   ePA_CHEVRON_UP,                        // Chevron Up
   ePA_CHEVRON_DOWN,                      // Chevron Down
   ePA_ARROW_EMPTY_CIRCLED_RIGHT,         // Arrow Empty Right Circled
   ePA_ARROW_EMPTY_CIRCLED_UP,            // Arrow Empty Up Circled
   ePA_ARROW_EMPTY_CIRCLED_DOWN,          // Arrow Empty Down Circled
   ePA_ARROW_RIGHT,                       // Arrow Right
   ePA_ARROW_UP,                          // Arrow Up
   ePA_ARROW_DOWN,                        // Arrow Down
   ePA_ARROW_UR,                          // Arrow Upper Right
   ePA_ARROW_DR,                          // Arrow Down Right
   ePA_ARROW_BOLD_RIGHT,                  // Arrow Right
   ePA_ARROW_BOLD_UP,                     // Arrow Bold Up
   ePA_ARROW_BOLD_DOWN,                   // Arrow Bold Down
   ePA_ARROW_BOLD_UR,                     // Arrow Bold Upper Right
   ePA_ARROW_BOLD_DR,                     // Arrow Bold Down Right
   ePA_ARROW_EMPTY_RIGHT,                 // Arrow Empty Right
   ePA_ARROW_EMPTY_UP,                    // Arrow Empty Up
   ePA_ARROW_EMPTY_DOWN,                  // Arrow Empty Down
   ePA_ARROW_EMPTY_UR,                    // Arrow Empty Upper Right
   ePA_ARROW_EMPTY_DR,                    // Arrow Empty Down Right
   ePA_X_MARK,                            // X Mark
   ePA_CHECK_MARK,                        // Check Mark
   ePA_X_MARK_BOX,                        // Box with X Mark
   ePA_CHECK_MARK_BOX                     // Box with Check Mark
};
class AlignableArrowAdjuster{		// Exactly Alignable arrows offset adjuster.
 public:
               AlignableArrowAdjuster(void)	 : mChartScale(0){}			//[CTor]
 public:                                                             // Methods
   PRICE       align_bot(AlignableArrow aPA, PRICE aPrice)	const{
		gsTicker.enable(1);	return aPrice + mPAD[aPA].mnUp * mChartScale;  		 }
   PRICE       align_hot(AlignableArrow aPA, PRICE aPrice) 	const{
		gsTicker.enable(1);  return aPrice + mPAD[aPA].mnHot * mChartScale; 		 }
   PRICE       align_top(AlignableArrow aPA, PRICE aPrice) 	const{
		gsTicker.enable(1);  return aPrice + mPAD[aPA].mnDn * mChartScale;  		 }
   Wingding    get_ac(AlignableArrow aPA)               		const{
		gsTicker.enable(1);  return mPAD[aPA].mArrowCode; 								 }
	void			on_timer(void){
		double   csPrev   = mChartScale;
			int      chp   = (int)ChartGetInteger(0, CHART_HEIGHT_IN_PIXELS);
																			if(chp == 0)   return;
		mChartScale = ( ChartGetDouble(0, CHART_PRICE_MAX)
						  - ChartGetDouble(0, CHART_PRICE_MIN)
						  ) / chp;
		if(mChartScale != csPrev){
				static int  NAspread[];    static long NAvolume[];
				/*(void)*/OnCalculate(Bars,         // rates_total,
											 0,            // prev_calculated,
											 Time, Open, High, Low, Close,
											 NAvolume,     // NU tick_volume[],
											 NAvolume,     // NU volume[],
											 NAspread);    // NU spread[]
		}
	}	// on_timer
private:                                                            // Data
	double  							mChartScale;	// [0.0]
	struct AlignableArrowData{	int	mnUp, mnDn, mnHot;	Wingding mArrowCode;	};
	const static AlignableArrowData	mPAD[];
};	// AlignableArrowAdjuster
const AlignableArrowData	AlignableArrowAdjuster::mPAD[]={		// Private kData
	{4,-4, 0,eWD_CIRCLE_SOLID},
	{2,-2, 0,eWD_CIRCLE_SOLID_MEDIUM},		{4,-4, 0,eWD_SQUARE_SOLID},
	{3,-3, 0,eWD_SQUARE_SOLID_SMALL},		{5,-6, 0,eWD_BOX},
	{5,-4, 0,eWD_DIAMOND_SOLID},				{6,-6, 0,eWD_RHOMBUS_SOLID},
	{6,-6,-1,eWD_RHOMBUS_EMPTY_X},			{6,-6, 0,eWD_INTEREST},
	{6,-6, 0,eWD_CIRCLE},						{6,-6, 0,eWD_CIRCLE_BOLD},
	{6,-6, 0,eWD_CIRCLE_HEAVY},				{6,-6, 0,eWD_FISHEYE},
	{6,-6, 0,eWD_BULLSEYE},						{2,-5, 0,eWD_STAR_3},
	{4,-5, 0,eWD_STAR_4},						{3,-4, 0,eWD_STAR_5},
	{4,-5, 0,eWD_STAR_6},						{4,-4, 0,eWD_STAR_8},
	{4,-5, 0,eWD_STAR_12},						{4,-5,-1,eWD_POSITION_SQUARE},
	{4,-5,-1,eWD_POSITION_CIRCLE},			{4,-4, 0,eWD_DIAMOND_CONCAVE},
	{6,-6, 0,eWD_CIRCLED_PENTASTAR},			{5,-6, 0,eWD_CHEVRON_RIGHT},
	{4,-6,-6,eWD_CHEVRON_UP},					{6,-4, 6,eWD_CHEVRON_DOWN},
	{6,-6, 0,eWD_ARROW_EMPTY_CIRCLED_RIGHT},{6,-6,-6,eWD_ARROW_EMPTY_CIRCLED_UP},
	{6,-6, 6,eWD_ARROW_EMPTY_CIRCLED_DOWN}, {5,-6, 0,eWD_ARROW_RIGHT},
	{6,-5,-5,eWD_ARROW_UP},						{5,-7, 5,eWD_ARROW_DOWN},
	{4,-5,-5,eWD_ARROW_UR},						{6,-6, 6,eWD_ARROW_DR},
	{6,-6, 0,eWD_ARROW_BOLD_RIGHT},			{7,-6,-6,eWD_ARROW_BOLD_UP},
	{5,-8, 5,eWD_ARROW_BOLD_DOWN},			{5,-5,-5,eWD_ARROW_BOLD_UR},
	{5,-6, 5,eWD_ARROW_BOLD_DR},				{5,-6,-1,eWD_ARROW_EMPTY_RIGHT},
	{6,-6,-6,eWD_ARROW_EMPTY_UP},				{5,-6, 5,eWD_ARROW_EMPTY_DOWN},
	{3,-6,-6,eWD_ARROW_EMPTY_UR},				{5,-3, 5,eWD_ARROW_EMPTY_DR},
	{3,-3, 0,eWD_X_MARK},						{3,-6, 0,eWD_CHECK_MARK},
	{6,-6, 0,eWD_X_MARK_BOX},					{6,-6, 0,eWD_CHECK_MARK_BOX} };// mPAD
AlignableArrowAdjuster	gsAAA;	// Singleton.
void     OnTimerAlignable(void){	gsAAA.on_timer();	}
//} Exactly alignable arrows.

// Predefined in MT5 {DRAW_XXXX}. In Mt4, same value as DRAW_XXXX
enum  ENUM_DRAW_TYPE {  TYPE_LINE,     TYPE_SECTION,     TYPE_HISTOGRAM,
                        TYPE_ARROW,    TYPE_ZIGZAG,      TYPE_NONE=12   };

// ENUM_LINE_STYLE {    STYLE_SOLID,   STYLE_DASH,       STYLE_DOT,
//                      STYLE_DASHDOT, STYLE_DASHDOTDOT  };
#define  ENUM_LINE_STYLE_COUNT 5
#define  INCREMENT_LINE_STYLE(S, N)                                           \
   ENUM_LINE_STYLE( (S + N + ENUM_LINE_STYLE_COUNT) % ENUM_LINE_STYLE_COUNT )

class BufferDefinition{
 public:
   void     BufferDefinition(COUNT nLine)  : miBuf(nLine-1),            // CTor
                                                             meDraw(DRAW_LINE){}
 public:                                                             // Methods
   const BufferDefinition*
            arr(AlignableArrow eAA){            return arr(gsAAA.get_ac(eAA)); }
   const BufferDefinition*
            arr(Wingding eAC){                  SetIndexArrow(miBuf,eAC);
                                                return typ(TYPE_ARROW).wid(1); }
   bool     buf(PRICE& bBuf[])            const{return
                  SetIndexBuffer(miBuf, bBuf) && ArraySetAsSeries(bBuf,true);  }
   const BufferDefinition*
            clr(color eColor)             const{
            SetIndexStyle(miBuf, meDraw, EMPTY, EMPTY, eColor);   return &this;}
   const BufferDefinition*
            ev(PRICE aEV)                 const{SetIndexEmptyValue(miBuf, aEV);
                                                                  return &this;}
   const BufferDefinition*
            lab(string label)             const{SetIndexLabel(miBuf, label);
                                                                  return &this;}
   // Buffer is not shown on the chart.
   // Prefer "#property   indicator_colorX  clrNONE" over noChart() so
   // property/color shows correct.
   const BufferDefinition*
            noChart(void)                 const{return clr(clrNONE);           }
   // Buffer is not shown in the data window.
   const BufferDefinition*
            noData(void)                  const{return lab(NULL);              }
   const BufferDefinition*
            shf(int aShift)               const{SetIndexShift(miBuf, aShift);
                                                                  return &this;}
   // #property   indicator_style1  STYLE_SOLID
   const BufferDefinition*
            sty(ENUM_LINE_STYLE aeStyle)  const{
                           SetIndexStyle(miBuf, meDraw, aeStyle); return &this;}
   // #property   indicator_type1   DRAW_LINE
   BufferDefinition* typ(ENUM_DRAW_TYPE aeDraw){         meDraw = aeDraw;
                                    SetIndexStyle(miBuf, meDraw); return &this;}
   const BufferDefinition*
            wid(COUNT aWidth)             const{
               SetIndexStyle(miBuf, meDraw, EMPTY, (int)aWidth);  return &this;}
 private:                                                            // Data
   INDEX          miBuf;   // Buffer index.
   ENUM_DRAW_TYPE meDraw;  // Buffer draw type.
}; // BufferDefinition
//+- indicator_utilities.mqh ----------------------------------------+
